package com.frist008.nixlibrary.database;

import android.content.ContentProvider;
import android.content.ContentResolver;
import android.content.ContentUris;
import android.content.ContentValues;
import android.content.Context;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteOpenHelper;
import android.net.Uri;
import android.support.annotation.IntDef;
import android.support.annotation.NonNull;
import android.support.annotation.Nullable;
import android.text.TextUtils;

import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

@SuppressWarnings("unused")
public abstract class NixProvider extends ContentProvider {

	static final NixUriMatcher URI_MATCHER = new NixUriMatcher();

	@NonNull
	@SuppressWarnings("NullableProblems")
	private static NixProvider provider;

	@Nullable
	private ContentResolver resolver;
	@NonNull
	@SuppressWarnings("NullableProblems")
	private final String authorities;
	@NonNull
	@SuppressWarnings("NullableProblems")
	private SQLiteDatabase database;
	@NonNull
	private final String dbName;
	private final int dbVersion;

	public NixProvider(@NonNull String authorities, int dbVersion) {
		this(authorities, NixProvider.class.getName().replace(".", ""), dbVersion);
	}

	public NixProvider(@NonNull String authorities, @NonNull String dbName, int dbVersion) {
		NixProvider.provider = this;
		this.authorities = authorities;
		this.dbName = dbName;
		this.dbVersion = dbVersion;
	}

	public static ContentResolver getContentResolver() {
		initContentResolver();
		return provider.resolver;
	}

	private static void initContentResolver() {
		Context context = provider.getContext();

		if (context != null) {
			ContentResolver resolver = context.getContentResolver();

			if (resolver != null) {
				provider.resolver = resolver;
			}
		}
	}

	@Override
	public boolean onCreate() {
		initContentResolver();
		initTableOrView(new NixTableOrViewInitializer());

		SQLiteOpenHelper databaseHelper = new NixSqlHelper(getContext(), dbName, dbVersion);
		this.database = databaseHelper.getWritableDatabase();
		return true;
	}

	@NonNull
	protected static NixProvider getProvider() {
		return provider;
	}

	@NonNull
	public static String getAuthorities() {
		return getProvider().authorities;
	}

	@Deprecated
	@Nullable
	@Override
	public String getType(@NonNull Uri uri) {
		List<String> pathSegmentUriList = uri.getPathSegments();
		boolean notTable = pathSegmentUriList.size() != 1;
		String nameTable = getNameTable(notTable, pathSegmentUriList);

		return getType(uri, nameTable);
	}

	@Nullable
	private String getType(@NonNull Uri uri, @NonNull String nameTable) {
		return null;
	}

	@Override
	public Cursor query(@NonNull Uri tableUri,
	                    @Nullable String[] projection,
	                    @Nullable String selection,
	                    @Nullable String[] selectionArgs,
	                    @Nullable String sortOrder) {
		ContentResolver resolver = getContentResolver();
		List<String> pathSegmentUriList = tableUri.getPathSegments();
		boolean notTable = pathSegmentUriList.size() != 1;
		String nameTable = getNameTable(notTable, pathSegmentUriList);
		Set<NixTableOrView> tableOrViewSet = URI_MATCHER.getTableUriSet(nameTable);
		selectionArgs = fixedIllegalArgumentException(selectionArgs);

		if (notTable) {
			selection = addToSelectLastPathSegment(selection, tableUri);
		}

		Cursor cursor = database.query(nameTable,
		                               projection,
		                               selection,
		                               selectionArgs,
		                               null,
		                               null,
		                               sortOrder);

		if (tableOrViewSet != null) {
			for (NixTableOrView tableOrView : tableOrViewSet) {
				cursor.setNotificationUri(resolver, tableOrView.getUri());
			}
		} else {
			cursor.setNotificationUri(resolver, tableUri);
		}

		return cursor;
	}

	private String[] fixedIllegalArgumentException(@Nullable String[] selectionArgs) {
		selectionArgs = selectionArgs != null && selectionArgs.length > 0 ? selectionArgs : null;

		if (selectionArgs != null) {
			for (int i = 0; i < selectionArgs.length; i++) {
				if (selectionArgs[i] == null) {
					selectionArgs[i] = NixWhere.NULL;
				}
			}
		}

		return selectionArgs;
	}

	@Override
	public int bulkInsert(@NonNull Uri tableUri, @NonNull ContentValues[] valuesArr) {
		List<String> pathSegmentUriList = tableUri.getPathSegments();
		boolean notTable = pathSegmentUriList.size() != 1;
		String nameTable = getNameTable(notTable, pathSegmentUriList);
		int conflictAlgorithm = getConflictAlgorithm(nameTable);

		try {
			database.beginTransaction();

			for (ContentValues values : valuesArr) {
				database.insertWithOnConflict(nameTable, null, values, conflictAlgorithm);
			}

			database.setTransactionSuccessful();
		} finally {
			database.endTransaction();
		}

		if (valuesArr.length > 0) {
			notifyChange(nameTable);
		}

		return valuesArr.length;
	}

	@Override
	public Uri insert(@NonNull Uri tableUri, @Nullable ContentValues values) {
		List<String> pathSegmentUriList = tableUri.getPathSegments();
		boolean notTable = pathSegmentUriList.size() != 1;
		String nameTable = getNameTable(notTable, pathSegmentUriList);
		long rowId = database.insertWithOnConflict(nameTable,
		                                           null,
		                                           values,
		                                           getConflictAlgorithm(nameTable));
		if (rowId > 0) {
			notifyChange(nameTable);
		}

		return ContentUris.withAppendedId(tableUri, rowId);
	}

	@ConflictAlgorithm
	protected abstract int getConflictAlgorithm(@NonNull String nameTable);

	@Override
	public int delete(@NonNull Uri tableUri,
	                  @Nullable String selection,
	                  @Nullable String[] selectionArgs) {
		List<String> pathSegmentUriList = tableUri.getPathSegments();
		boolean notTable = pathSegmentUriList.size() != 1;
		String nameTable = getNameTable(notTable, pathSegmentUriList);
		selectionArgs = fixedIllegalArgumentException(selectionArgs);

		if (notTable) {
			selection = addToSelectLastPathSegment(selection, tableUri);
		}

		int count = database.delete(nameTable, selection, selectionArgs);

		if (count > 0) {
			notifyChange(nameTable);
		}

		return count;
	}

	@Override
	public int update(@NonNull Uri tableUri,
	                  @Nullable ContentValues values,
	                  @Nullable String selection,
	                  @Nullable String[] selectionArgs) {
		List<String> pathSegmentUriList = tableUri.getPathSegments();
		boolean notTable = pathSegmentUriList.size() != 1;
		String nameTable = getNameTable(notTable, pathSegmentUriList);
		selectionArgs = fixedIllegalArgumentException(selectionArgs);

		if (notTable) {
			selection = addToSelectLastPathSegment(selection, tableUri);
		}

		int count = database.updateWithOnConflict(nameTable,
		                                          values,
		                                          selection,
		                                          selectionArgs,
		                                          getConflictAlgorithm(nameTable));
		if (count > 0) {
			notifyChange(nameTable);
		}

		return count;
	}

	public static void notifyChange(@NonNull NixView view) {
		getContentResolver().notifyChange(view.getUri(), null);
		getProvider().notifyChange(view.toString());
	}

	public static void notifyChange(@NonNull NixView... views) {
		for (NixView view : views) {
			notifyChange(view);
		}
	}

	public static void notifyChange(@NonNull NixTable table) {
		(getProvider()).notifyChange(table.toString());
	}

	public static void notifyChange(@NonNull NixTable... tables) {
		for (NixTable table : tables) {
			notifyChange(table);
		}
	}

	public static void notifyChangeAll(@Nullable NixTable... tableFilter) {
		searchAllTables(new SearchManager() {
			@Override
			public void searchComplete(@NonNull NixTable table) {
				notifyChange(table);
			}
		}, tableFilter);
	}

	private void notifyChange(@Nullable String nameTable) {
		Set<NixTableOrView> tableOrViewSet = URI_MATCHER.getTableUriSet(nameTable);
		ContentResolver resolver = getContentResolver();

		if (tableOrViewSet != null) {
			for (NixTableOrView tableOrView : tableOrViewSet) {
				resolver.notifyChange(tableOrView.getUri(), null);
			}
		}
	}

	private String getNameTable(boolean isNotTable, @NonNull List<String> pathSegmentUriList) {
		return isNotTable
		       ? pathSegmentUriList.get(pathSegmentUriList.size() - 2)
		       : pathSegmentUriList.get(pathSegmentUriList.size() - 1);
	}

	private String addToSelectLastPathSegment(@Nullable String selection, @NonNull Uri uri) {
		selection = ((selection == null) ? "" : selection);

		return (TextUtils.isEmpty(selection) ? "" : (selection + NixWhere.AND))
		       + NixWhere.NixCols.COLUMN_ID + NixWhere.EQUALS + uri.getLastPathSegment();
	}

	protected void onCreateDB(@NonNull SQLiteDatabase sqLiteDatabase) {
		Set<NixView> nixViewSet = new HashSet<>();

		for (Set<NixTableOrView> tableOrViewSet : URI_MATCHER.values()) {
			for (NixTableOrView tableOrView : tableOrViewSet) {
				if (tableOrView instanceof NixView) {
					findView(nixViewSet, (NixView) tableOrView);
				} else {
					autoCreationTable(sqLiteDatabase, tableOrView);
				}
			}
		}

		for (NixView nixView : nixViewSet) {
			sqLiteDatabase.execSQL(nixView.getQueryForCreation());
		}
	}

	private void findView(Set<NixView> nixViewSet, NixView tableOrView) {
		if (tableOrView.isNotDeprecated()) {
			nixViewSet.add(tableOrView);
		}
	}

	private void autoCreationTable(@NonNull SQLiteDatabase sqLiteDatabase,
	                               NixTableOrView tableOrView) {
		NixTable table = (NixTable) tableOrView;

		if (table.isNotDeprecated()) {
			sqLiteDatabase.execSQL(table.getQueryForCreation());
			Set<String> indexColumnSet = table.getQueryForCreationIndex();

			for (String indexColumn : indexColumnSet) {
				sqLiteDatabase.execSQL(NixTable.createIndex(table, indexColumn));
			}
		}
	}

	protected void onUpdateDB(@NonNull SQLiteDatabase sqLiteDatabase,
	                          int oldVersion,
	                          int newVersion) {

	}

	private class NixSqlHelper extends SQLiteOpenHelper {

		public NixSqlHelper(@Nullable Context context, @NonNull String name, int version) {
			super(context, name, null, version);
		}

		@Override
		public void onCreate(@NonNull SQLiteDatabase sqLiteDatabase) {
			onCreateDB(sqLiteDatabase);
		}

		@Override
		public void onUpgrade(@NonNull SQLiteDatabase sqLiteDatabase,
		                      int oldVersion,
		                      int newVersion) {
			onUpdateDB(sqLiteDatabase, oldVersion, newVersion);
		}
	}

	@Retention(RetentionPolicy.SOURCE)
	@IntDef({SQLiteDatabase.CONFLICT_ABORT,
	         SQLiteDatabase.CONFLICT_FAIL,
	         SQLiteDatabase.CONFLICT_IGNORE,
	         SQLiteDatabase.CONFLICT_NONE,
	         SQLiteDatabase.CONFLICT_REPLACE,
	         SQLiteDatabase.CONFLICT_ROLLBACK})
	protected @interface ConflictAlgorithm {

	}

	// query

	@SuppressWarnings("SuspiciousMethodCalls")
	public static void clearAllTable(@Nullable NixTable... tableFilter) {
		searchAllTables(new SearchManager() {
			@Override
			public void searchComplete(@NonNull NixTable table) {
				getContentResolver().delete(table.getUri(), null, null);
			}
		}, tableFilter);
	}

	@SuppressWarnings("SuspiciousMethodCalls")
	public static void searchAllTables(@NonNull SearchManager searchManager,
	                                   @Nullable NixTable... nameTableFilter) {
		List<NixTable> nameList = nameTableFilter == null
		                          ? new ArrayList<NixTable>()
		                          : Arrays.asList(nameTableFilter);

		for (Set<NixTableOrView> tableOrViewSet : URI_MATCHER.values()) {
			for (NixTableOrView tableOrView : tableOrViewSet) {
				if (!nameList.contains(tableOrView)
				    && tableOrView instanceof NixTable
				    && tableOrView.isNotDeprecated()) {

					try {
						searchManager.searchComplete((NixTable) tableOrView);
					} catch (Exception ignore) {
					}
				}
			}
		}
	}

	protected abstract void initTableOrView(@NonNull NixTableOrViewInitializer initializer);

	public interface SearchManager {
		void searchComplete(@NonNull NixTable table);
	}
}
