package com.frist008.nixlibrary.database;

import android.content.ContentValues;
import android.net.Uri;
import android.support.annotation.NonNull;
import android.support.annotation.Nullable;

import java.text.MessageFormat;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

@SuppressWarnings("unused")
public class NixTable<Bean extends NixTableBean<Bean>> extends NixTableOrView {

	private static final String NOT_NULL = " NOT NULL ";
	private static final String DEFAULT = " DEFAULT ";
	private static final String UNIQUE = " UNIQUE ";

	protected static final String CREATE_INDEX = "CREATE INDEX IF NOT EXISTS {0} ON {1} ({2})";
	private static final String CREATE_TABLE = "CREATE TABLE IF NOT EXISTS {0} ({1});";
	protected static final String DROP_TABLE = "DROP TABLE IF EXISTS {0};";
	protected static final String DROP_VIEW = "DROP VIEW IF EXISTS {0};";

	private final Uri contentUri;
	private final Bean bean;

	private final Set<String> indexColumnSet = new HashSet<>();

	public NixTable(Bean bean) {
		this(bean, false);
	}

	public NixTable(Bean bean, boolean deprecated) {
		this(bean, bean.getClass().getSimpleName(), deprecated);
	}

	public NixTable(Bean bean, @NonNull String nameTable) {
		this(bean, nameTable, false);
	}

	public NixTable(Bean bean, @NonNull String nameTable, boolean deprecated) {
		super(nameTable, deprecated);
		this.bean = bean;
		this.contentUri = NixProvider.URI_MATCHER.newUriTable(this);
	}

	@Override
	public Uri getUri() {
		return contentUri;
	}

	public Bean getBean() {
		return bean;
	}

	@Override
	public String getQueryForCreation() {
		StringBuilder builder = new StringBuilder();
		fillFields(builder);
		return createTable(this, builder);
	}

	private void fillFields(StringBuilder builder) {
		Collection<NixColumnType> columnCollection = bean.columnMap.values();
		for (NixColumnType column : columnCollection) {
			if (column.isIndex()) {
				indexColumnSet.add(column.toString());
			}

			if (builder.length() > 0) {
				builder.append(NixWhere.COMMA);
			}

			builder.append(column).append(column.getType());

			if (column.isUnique()) {
				builder.append(UNIQUE);
			}

			if (column.isNotNull()) {
				builder.append(NOT_NULL);
			}

			if (column.hasDefValue()) {
				builder.append(DEFAULT).append(column.getDBDefValue());
			}
		}
	}

	public Set<String> getQueryForCreationIndex() {
		return indexColumnSet;
	}

	protected static String createTable(NixTable table, StringBuilder tableBuilder) {
		return MessageFormat.format(CREATE_TABLE, table, tableBuilder);
	}

	protected static String createIndex(NixTable table, String column) {
		return MessageFormat.format(CREATE_INDEX, table + column, table, column);
	}

	public <Bean extends NixTableBean> int insert(@NonNull List<Bean> beanCollection) {
		return NixTableBean.insert(beanCollection);
	}

	private int deleteRow(@NonNull Uri url,
	                      @Nullable String where,
	                      @Nullable String... selectionArgs) {
		return NixProvider.getContentResolver().delete(url, where, selectionArgs);
	}

	public int deleteRow(@Nullable String where, @Nullable String... selectionArgs) {
		return deleteRow(getUri(), where, selectionArgs);
	}

	public int deleteRow(long id) {
		return deleteRow(addLastSegment(id), null);
	}

	public int deleteAll() {
		return deleteRow(null);
	}

	private int updateRow(@NonNull Uri url,
	                      @NonNull ContentValues cv,
	                      @Nullable String where,
	                      @Nullable String... selectionArgs) {
		return NixProvider.getContentResolver().update(url, cv, where, selectionArgs);
	}

	public int updateRow(@NonNull ContentValues cv, String where, String... selectionArgs) {
		return updateRow(getUri(), cv, where, selectionArgs);
	}

	public int updateRow(long id, @NonNull ContentValues cv) {
		return updateRow(addLastSegment(id), cv, null);
	}

	public int updateAll(@NonNull ContentValues cv) {
		return updateRow(cv, null);
	}
}
